/* BLOB.CPP - Programmed by Maarten Hofman */
/* 100404: Started programming */
/* 140404: Added heartbeat for fluent movement */
/* 180404: Adding angles */
/* 190404: Fixed animation stuff */
/* 220404: Add special animation stuff */
/* 260404: Added the has_base_name() functionality */
/* 280404: move_coordinates() sets moved */
/* 060504: Added rotation function */
/* 040704: Added transparancy indicator */
/* 150704: Made rotation more smooth */
/* 240704: Added get_parent option */
/* 020904: Added fluid rotate */
/* 030904: Added y rotate option */
/* 130904: Tried to slow down movement a bit */
/* 111004: Added speed parameter */
/* 281004: Added parameter to do_rotate */
/* 011104: Made it possible to increase the speed of movement */
/* 230205: Added set angle */
/* 260205: Added rotation to body */
/* 270205: Made it possible to set the copy regardless */

#include <string.h>
#include <stdio.h>
#include <malloc.h>
#include "blob.h"
#include "ps2quaternion.h"

using namespace KazMaths;

blob::blob() {
  cx=-1;
  cy=-50;
  cz=-1;
  tx=-2;
  ty=-2;
  tz=-2;
  loc=new Matrix4x4();
  parent=NULL;
  step=0;
  animation=NULL;
  ax=0;
  ay=0;
  az=0;
  bias=0;
  rotate=0;
  animate=false;
  updated=false;
  moved=false;
  angle=0;
  ta=0;
  trans=false;
  rotx=0;
  roty=0;
  rotz=0;
  crotx=0;
  croty=0;
  crotz=0;
  copied=false;
  cupdated=false;
  animspeed=1;
  movespeed=1;
}

void blob::set_bias() {
  bias=0.1;
}

void blob::set_updated() {
  if(parent) parent->set_updated();
  cupdated=true;
}

bool blob::get_updated() {
  if(parent) return parent->get_updated();
  return cupdated;
}

bool blob::is_trans() {
  return trans;
}

void blob::set_trans() {
  trans=true;
}

void blob::set_animate() {
  animate=true;
  fprintf(stderr,"Animating blob\n");
}

void blob::set_rotate(int how) {
  fprintf(stderr,"Rotate being set to %d in %s\n",how,name);
  rotate=how;
}

blob::~blob() {
  if(!copied) if(NULL!=animation) free(animation);
  delete loc;
}

void blob::set_animation_type(int how,int speed) {
  animtype=how;
  animspeed=speed;
  fprintf(stderr,"Animspeed set to: %d\n",animspeed);
}

void blob::set_animation(float *how) {
  animation=how;
}

void blob::set_copy(blobcopy *what) {
  what->parent=parent;
  if(parent==NULL) return;
  what->ax=ax;
  what->ay=ay;
  what->az=az;
  what->animtype=animtype;
  what->animation=animation;
  what->animspeed=animspeed;
  what->angle=angle;
  copied=true;
}

void blob::set_parent(blob *who) {
  parent=who;
}

bool blob::get_moved() {
  return moved;
}

void blob::set_relative(float x,float y,float z) {
  if(parent==NULL) return;
  ax=x;
  ay=y;
  az=z;
  cx=parent->get_cx()+ax;
  cy=parent->get_cy()+ay;
  cz=parent->get_cz()+az;
  loc->Translation(cx,cy+bias,cz);
  return;
}

bool blob::move_coordinates(int x,int y,int z) {
  if((cy<-45)&&(ty==-2)) {
    /* First time, so we are placed on the map */
    loc->Translation(x,y+bias,z);
    cx=x;
    cy=y;
    cz=z;
    tx=x;
    ty=y;
    tz=z;
    fprintf(stderr,"Placing at %d %d %d\n",x,y,z);
    return true;
  }
  fprintf(stderr,"Going to %d %d %d\n",x,y,z);
  tx=x;
  ty=y;
  tz=z;
  return false;
}

int blob::get_texture() {
  return texture;
}

int blob::get_model() {
  return model;
}

void blob::set_texture(int how) {
  texture=how;
}

void blob::set_model(int how) {
  model=how;
}

void blob::set_name(char *how) {
  strncpy(name,how,32);
}

int blob::has_name(char *how) {
  return !strcmp(name,how);
}

int blob::has_base_name(char *how) {
int match;

  match=!strncmp(name,how,strlen(how));
  if(match) {
    printf("Base name %s matched %s\n",how,name);
  }
  return match;
}

Matrix4x4 *blob::get_location() {
  return(loc);
}

float blob::get_cx() {
  return cx;
}

float blob::get_cy() {
  return cy;
}

float blob::get_cz() {
  return cz;
}

float blob::get_angle() {
  return angle;
}

void blob::set_angle(float how) {
Quaternion rotQuat;
Matrix4x4 matRot;

  angle=how;
  rotQuat.EulerAngles(0,how,0);
  matRot=rotQuat.ToRotationMatrix4x4();
  *loc=matRot*(*loc);
}

void blob::update() {
Quaternion rotQuat;
Matrix4x4 matRot;

  /* Parents don't need to update themselves */
  if(NULL==parent) return;
  cx=parent->get_cx()+ax;
  cy=parent->get_cy()+ay;
  cz=parent->get_cz()+az;
  angle=parent->get_angle();
  loc->Translation(cx,cy+bias,cz);
  if(animation==NULL) return;
  /* We only need the animation once */
  if(updated) return;
  updated=true;
  fprintf(stderr,"Rotating %s\n",name);
  switch(animtype) {
    case 257:		// Rotate around the Z axis
      rotQuat.EulerAngles(0,angle,animation[0]);
      matRot=rotQuat.ToRotationMatrix4x4();
      *loc=matRot*(*loc);
      break;
    case 258:		// Rotate around the X axis
      rotQuat.EulerAngles(animation[0],angle,0);
      matRot=rotQuat.ToRotationMatrix4x4();
      *loc=matRot*(*loc);
      break;
    case 259:
      rotQuat.EulerAngles(0,angle+animation[0],0);
      matRot=rotQuat.ToRotationMatrix4x4();
      *loc=matRot*(*loc);
      break;
    default:
      rotQuat.EulerAngles(0,angle,0);
      matRot=rotQuat.ToRotationMatrix4x4();
      *loc=matRot*(*loc);
      break;
  }
}

inline float blob::interpolate_animation() {
int loc;
float olda,newa;

  loc=step/animspeed;
  olda=animation[loc];
  newa=(loc==9)?animation[0]:animation[loc+1];
  return olda+((newa-olda)/(float)animspeed)*(float)(step%animspeed);
}

bool blob::heart_beat() {
Quaternion rotQuat;
Matrix4x4 matRot;
float a,p;
bool rotated;

  /* Are we completely initialised? */
  if((cy<-45)&&(ty==-2)) return false;
  if(parent) {
    moved=parent->get_moved();
    /* Check whether this is a special animation type */
    if((animtype&256)&&(animate)) moved=true;
    if(moved) {
      angle=parent->get_angle();
      update();
      /* Filter out the special bits of the animation type */
      switch(animtype&255) {
        case 1:		// Rotate around the Z axis
          if(NULL==animation) {
            printf("No animation available!\n");
            break;
          }
          rotQuat.EulerAngles(0,angle,interpolate_animation());
          matRot=rotQuat.ToRotationMatrix4x4();
          *loc=matRot*(*loc);
          break;
        case 2:		// Rotate around the X axis
          if(NULL==animation) {
            printf("No animation available!\n");
            break;
          }
          rotQuat.EulerAngles(interpolate_animation(),angle,0);
          matRot=rotQuat.ToRotationMatrix4x4();
          *loc=matRot*(*loc);
          break;
        case 3:		// Rotate around the Y axis
          if(NULL==animation) break;
          rotQuat.EulerAngles(0,angle+interpolate_animation(),0);
          matRot=rotQuat.ToRotationMatrix4x4();
          *loc=matRot*(*loc);
          break;
      }
      step++;
      if((animtype&256)&&(!animate)) step=0;
      if(step>=10*animspeed) {
        step=0;
        if(animate) {
          animate=false;
          printf("End of animation\n");
        }
      }
    }
//    else {
//      /* Synchronize animation */
//      step=0;
//    }
    return moved; 
  }
  rotated=false;
  /* Rotate checks */
  if(rotx-ROTATE_RESOLUTION>crotx) {
    crotx+=ROTATE_RESOLUTION;
    rotated=true;
  }
  if(rotx+ROTATE_RESOLUTION<crotx) {
    crotx-=ROTATE_RESOLUTION;
    rotated=true;
  }
  if(roty-ROTATE_RESOLUTION>croty) {
    croty+=ROTATE_RESOLUTION;
    rotated=true;
  }
  if(roty+ROTATE_RESOLUTION<croty) {
    croty-=ROTATE_RESOLUTION;
    rotated=true;
  }
  if(rotz-ROTATE_RESOLUTION>crotz) {
    crotz+=ROTATE_RESOLUTION;
    rotated=true;
  }
  if(rotz+ROTATE_RESOLUTION<crotz) {
    crotz-=ROTATE_RESOLUTION;
    rotated=true;
  }
  if(rotated) internal_do_rotate(crotx,croty,crotz);
  /* Move code */
  p=3.14;
  a=p*4;
  moved=false;
  if(ty-MOVE_RESOLUTION>cy) {
    cy+=MOVE_RESOLUTION;
    moved=true;
  }
  if(ty+MOVE_RESOLUTION<cy) {
    cy-=MOVE_RESOLUTION;
    moved=true;
  }
  if(tx-MOVE_RESOLUTION>cx) {
    cx+=MOVE_RESOLUTION;
    moved=true;
    a=p/2;
  }
  if(tx+MOVE_RESOLUTION<cx) {
    cx-=MOVE_RESOLUTION;
    moved=true;
    a=-p/2;
  }
  if(tz-MOVE_RESOLUTION>cz) {
    cz+=MOVE_RESOLUTION;
    moved=true;
    if(a<3*p) {
      a=a/2;
    }
    else {
      a=0;
    }
  }
  if(tz+MOVE_RESOLUTION<cz) {
    cz-=MOVE_RESOLUTION;
    moved=true;
    if(a<3*p) {
      a=a/2+a;
    }
    else {
      a=p;
    }
  }
  if(rotate==2) {
    loc->Translation(cx,cy+bias,cz); 
    rotQuat.EulerAngles(0,angle,0);
    matRot=rotQuat.ToRotationMatrix4x4();
    *loc=matRot*(*loc);
    angle+=0.1;
    if(angle>2*p) angle=0;
    return moved;
  }
  if(moved) {
    loc->Translation(cx,cy+bias,cz); 
    if(rotate==1) {
      if(a<3*p) {
        ta=a;
        if(angle-ta>p) angle-=2*p;
        if(angle-ta<-p) angle+=2*p;
        if(angle<ta-0.1) angle+=0.1;
        if(angle>ta+0.1) angle-=0.1;
        rotQuat.EulerAngles(0,angle,0);
        matRot=rotQuat.ToRotationMatrix4x4();
        *loc=matRot*(*loc);
      }
    }
  }
  return moved;
}

void blob::internal_do_rotate(float x,float y,float z) {
Quaternion rotQuat;
Matrix4x4 matRot;
 
  loc->Translation(cx,cy+bias,cz); 
  rotQuat.EulerAngles(x,y,z);
  matRot=rotQuat.ToRotationMatrix4x4();
  *loc=matRot*(*loc);
}

void blob::do_rotate(float x,float y,float z,int how) {
  rotx=x;
  roty=y;
  rotz=z;
  if(!how) return;
  crotx=x;
  croty=y-2*ROTATE_RESOLUTION;
  crotz=z;
}

float blob::target_distance() {
  return (tx-cx)*(tx-cx)+(ty-cy)*(ty-cy)+(tz-cz)*(tz-cz);
}

void blob::set_move_speed(int how) {
  movespeed=how;
}

